#--------------------------------
#	hashtrie.tcl
#	by Shinsuke Takeuchi
#	www.ctrans.org
#--------------------------------
package require Tcl 8.5
package provide hashtrie 1.0
#--------------------------------
namespace eval hashtrie {}
#--------------------------------
#	hashtrie::new
#--------------------------------
proc hashtrie::new {} {
	set goto ""; #dict for state transition(key & state)
	set output ""; #dict for replace datas(state & word)
	set unum 0; #unique number for state(integer)
	set org ""; #dict for original datas(state & word)
	return [list $goto $output $unum $org]
}
#--------------------------------
#	hashtrie::add
#--------------------------------
proc hashtrie::add {trie word value} {
	lassign $trie goto output unum org
	set state 0
	set flg 0
	set len [string length $word]
	#check $state and $key while $goto is defined
	for {set i 0} {$i < $len} {incr i} {
		#key format = char-state
		set key "[string index $word $i]-$state"
		if {[dict exists $goto $key]} {
			set state [dict get $goto $key]
		} else {
			set flg 1
			break
		}
	}
	#current word shorter than any words that were defined before,
	#so define new $output & $org now.
	if {$flg == 0} {
		dict set output $state $value
		dict set org $state $word
		return [list $goto $output $unum $org]
	}
	#define new $goto.
	for {set j $i} {$j < $len} {incr j} {
		incr unum
		set key "[string index $word $j]-$state"
		dict set goto $key $unum
		set state $unum
	}
	dict set output $state $value
	dict set org $state $word
	return [list $goto $output $unum $org]
}
#--------------------------------
#	hashtrie::remove
#	remove word from trie.
#--------------------------------
proc hashtrie::remove {trieName word} {
	upvar $trieName trie
	lassign $trie goto output unum org
	set matches [hashtrie::getall $trie $word]
	if {$matches == ""} {return 0}
	lassign [lindex $matches end] match len
	#update $output and $org.
	set output [dict remove $output $match]
	set org [dict remove $org $match]
	set trie [list $goto $output $unum $org]
	return 1
}
#--------------------------------
#	hashtrie::exists
#--------------------------------
proc hashtrie::exists {trie word} {
	lassign $trie goto output unum org
	set state 0
	set len [string length $word]
	for {set i 0} {$i < $len} {incr i} {
		set key "[string index $word $i]-$state"
		if {[dict exists $goto $key]} {
			set state [dict get $goto $key]
		} else {
			return 0
		}
	}
	if {![dict exists $output $state]} {
		return 0
	}
	return 1
}
#--------------------------------
#	hashtrie::commonprefix
#	common prefix search.
#	This proc will return list of {state, end_pos}.
#--------------------------------
proc hashtrie::commonprefix {trie str} {
	lassign $trie goto output unum org
	set state 0
	set matches [list]
	set len [string length $str]
	for {set i 0} {$i < $len} {incr i} {
		set key "[string index $str $i]-$state"
		if {[dict exists $goto $key]} {
			set state [dict get $goto $key]
			if {[dict exists $output $state]} {
				lappend matches [list $state $i]
			}
		} else {
			break
		}
	}
	return $matches
}
#--------------------------------
#	hashtrie::longest
#	find longest word in string that is included in trie.
#	This proc will return {state end_pos} or blank.
#--------------------------------
proc hashtrie::longest {trie str} {
	set matches [hashtrie::commonprefix $trie $str]
	return [lindex $matches end]
}
#--------------------------------
#	hashtrie::getall
#	get all longest words that included in $str.
#	This proc will return list of {state start_pos end_pos}.
#--------------------------------
proc hashtrie::getall {trie str} {
	lassign $trie goto output unum org
	set state 0
	set result [list]
	set match ""
	set len [string length $str]
	for {set i 0} {$i < $len} {incr i} {
		if {$state == 0} {
			set start $i
			if {$match != ""} {
				lappend result $match
				set match ""
			}
		}
		set key "[string index $str $i]-$state"
		if {[dict exists $goto $key]} {
			set state [dict get $goto $key]
			if {[dict exists $output $state]} {
				#We will get longest word because $match will be replaced by new word.
				set match [list $state $start [expr $i+1]]
			}
		} else {
			if {$state != 0} {
				incr i -1
				set state 0
			}
		}
	}
	if {$match != ""} {
		lappend result $match
	}
	return $result
}
#--------------------------------
#	hashtrie::getvalue
#	get value of $output that corresponding to $state.
#--------------------------------
proc hashtrie::getvalue {trie state} {
	lassign $trie goto output unum org
	if {[dict exists $output $state]} {
		return [dict get $output $state]
	}
	return
}
#--------------------------------
#	hashtrie::getoriginal
#	get value of $org that corresponding to $state.
#--------------------------------
proc hashtrie::getoriginal {trie state} {
	lassign $trie goto output unum org
	if {[dict exists $org $state]} {
		return [dict get $org $state]
	}
	return
}
#--------------------------------
#	hashtrie::replace
#	search longest words that is included in trie from $str,
#	and replace to new word that is defined in $output.
#--------------------------------
proc hashtrie::replace {trie str {appendMode append} {template %s} {min 0}} {
	lassign $trie goto output unum org
	set new ""
	while {$str !=""} {
		#longest common prefix search.
		set state 0
		set len [string length $str]
		set match_state ""
		set match_len ""
		for {set i 0} {$i < $len} {incr i} {
			set key "[string index $str $i]-$state"
			if {[dict exists $goto $key]} {
				set state [dict get $goto $key]
				if {[dict exists $output $state]} {
					set match_state $state
					set match_len $i
				}
			} else {
				break
			}
		}
		#appendMode : lappend or append
		if {$match_state != ""} {
			if {$match_len <= $min} {
				$appendMode new [dict get $output $match_state]
			} else {
				$appendMode new [format $template [dict get $output $match_state]]
			}
			set str [string range $str [incr match_len] end]
		} else {
			$appendMode new [string index $str 0]
			set str [string range $str 1 end]
		}
	}
	return $new
}
#--------------------------------
#	hashtrie::import
#	import datas( key & value ) from dict.
#	mode:ignore, merge, update
#	mark:additional data for $org(ex.for export user data).
#	widget: a variable for showing progress(ex.label widget)
#--------------------------------
proc hashtrie::import {trie dict {mode update} {mark {}} {widget {}}} {
	if {$widget != ""} {
		upvar $widget progress
	}
	lassign $trie goto output unum org
	if {$unum == ""} {set unum 0}
	set count 0
	dict for {word value} $dict {
		incr count
		set progress "Progress: ${count} datas..."
		if {$word == ""} {continue}
		set state 0
		set flg 0
		set len [string length $word]
		#check $state and $key while $goto is defined.
		for {set i 0} {$i < $len} {incr i} {
			set key "[string index $word $i]-$state"
			if {[dict exists $goto $key]} {
				set state [dict get $goto $key]
			} else {
				set flg 1
				break
			}
		}
		#current word shorter than any words that were defined before,
		#so define new $output & $org now.
		if {$flg == 0} {
			#mode:ignore, merge, update
			if {[dict exists $output $state]} {
				if {$mode == "ignore"} {continue}
				if {$mode == "merge"} {
					set current [dict get $output $state]
					lappend value {*}$current
					set value [lsort -unique $value]
				}
			}
			dict set output $state $value
			if {$mark != ""} {
				dict set org $state "${mark}${word}"
			} else {
				dict set org $state $word
			}
			continue
		}
		#define new $goto.
		for {set j $i} {$j < $len} {incr j} {
			incr unum
			set key "[string index $word $j]-$state"
			dict set goto $key $unum
			set state $unum
		}
		dict set output $state $value
		if {$mark != ""} {
			dict set org $state "${mark}${word}"
		} else {
			dict set org $state $word
		}
	}
	return [list $goto $output $unum $org]
}
#--------------------------------
#	hashtrie::restore 
#	export original datas( key & value ) from trie.
#--------------------------------
proc hashtrie::export {trie {mark {}}} {
	lassign $trie goto output unum org
	set result [dict create]
	dict for {key original} $org {
		if {$mark != ""} {
			if {[string first $mark $original] != 0} {continue}
			set len [string length $mark]
			set original [string range $original $len end]
			set out [dict get $output $key]
			dict set result $original $out
		} else {
			set out [dict get $output $key]
			dict set result $original $out
		}
	}
	return $result
}
#--------------------------------
#	hashtrie::load 
#	load trie from file.
#--------------------------------
proc hashtrie::load {filename {enc utf-8} {nl auto}} {
	if {![file readable $filename]} {return}
	set f [open $filename r]
	fconfigure $f -encoding $enc -translation $nl
	set trie [read $f]
	close $f
	if {[string index $trie 0] == "\ufeff"} {
		set trie [string range $trie 1 end]
	}
	return $trie
}
#--------------------------------
#	hashtrie::save
#	save trie to file.
#--------------------------------
proc hashtrie::save {trie filename {enc utf-8} {nl auto}} {
	set f [open $filename w]
	fconfigure $f -encoding $enc -translation $nl
	if {$enc == "utf-8" || $enc == "unicode"} {
		puts -nonewline $f "\ufeff"
	}
	puts $f $trie
	close $f
	return 1
}